///////////////////////////////////////////
// Busy soft // LzxPack 01 // 02.02.2017 //
///////////////////////////////////////////

// Boundary for all file adress dependent debug outputs
#define ENAZAC 0
#define ENAEND 65535

// Debug output level for used types of compressions: 0=none, 1=basic, 2=enhanced, 3=brutal...
#define ENAINF 0

// Debug output level for LZ compression engine: 0=none, 1=small, 2=large, 3=oops, 4=WTF?!, 5=... //
#define ENACOM 0

// Enable debug output of reversing sekvences
#define ENAREV 0

// Enable debug output of show pack model
#define ENADSP 0

// Enable debug output of optimizing pack model
#define ENAOPT 0

// Enable debug output of generating output data
#define ENAOUT 0

// Enable debug output of using sekvences for processing pack type
//   UNP = first unpack tax,
//   PRV = direct estimation best sekvence,
//   CPT = capture sekvence,
//   SUM = final summary
#define ENAUNP 0
#define ENAPRV 0
#define ENACPT 0
#define ENASUM 0

// Select order number 0..N and name of compression for special displaying of sekvence selection (-1 = none)
#define SEKNUM -1
#define SEKNAM "LZX"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXLEN 65536
#define PACKTYPES 380
#define DEPACKERS 128

int * tstalloc(int num);

int index;
int bits_stream;
int poke_stream;
int packed_index;
int packed_bytes;
int unpack_bytes;
int overhead_bits;
int compressions;
int depackers;

int compress_type;
int compress_posi;
int offset1bits;
int offset2bits;

char name_out[1024];

int depacker_com[DEPACKERS];		// Depacker compress type
int depacker_pos[DEPACKERS];		// Depacker offset coding
int depacker_len[DEPACKERS];		// Depacker length

unsigned char input_data[MAXLEN];
unsigned char packed_data[MAXLEN << 2];

int hash_table[0x10000];	// Hash table for quick searching of words

int repper[MAXLEN];			// Period of repeated packed sequence ended in this byte in previous matches
int replen[MAXLEN];			// Length of repeated packed sequence ended in this byte in previous matches
int lstidx[MAXLEN];			// Locate of last processed sekvence matching with this position
int lstlen[MAXLEN];			// Length of last processed sekvence matching with this position
int bufnxt[MAXLEN];			// Lists of all found words - index of previous matching word

int typcom[PACKTYPES];		// Type of compression - coding length and unpacked data
int typpos[PACKTYPES];		// Type of compression - coding offsets
int declen[PACKTYPES];		// Depacked length (read from file in option -d)
int maxlen[PACKTYPES];		// Maximal length of sekvence for compression
int maxoff[PACKTYPES];		// Maximal offset of sekvence for compression
int ofset1[PACKTYPES];		// Bit width of offset 1 for LZX compression (lower value)
int ofset2[PACKTYPES];		// Bit width of offset 2 for LZX compression (higher value)
int locskp[PACKTYPES];		// Flag for skipping compression if best sekvence already found or offset is out of range
int loccnt[PACKTYPES];		// Counter of local unpack bytes (for evaluating of price of unpacked bytes)
int*limoff[PACKTYPES];		// Limits for offset
int*taxoff[PACKTYPES];		// Taxes  for offset
int*limlen[PACKTYPES];		// Limits for length
int*taxlen[PACKTYPES];		// Taxes  for length
int idxoff[PACKTYPES];		// Index into limoff/taxoff for actual position in input data
int idxlen[PACKTYPES];		// Index into limlen/taxlen for actual position in input data
int preoff[PACKTYPES];		// Index into limoff/taxoff for previous position in input data
int prelen[PACKTYPES];		// Index into limlen/taxlen for previous position in input data
int sumcom[PACKTYPES];		// Total length of packed data after LZ compression and adding end mark
int sumall[PACKTYPES];		// Total length of packed data after optimalization
int sumdec[PACKTYPES];		// Total length of packed data after optimalization and with depacker length
int sumpck[PACKTYPES];		// Number of saved bytes - bytes copied by sekvences
int sumunp[PACKTYPES];		// Number of unpacked bytes - bytes outside of sekvences
int sumove[PACKTYPES];		// Overhead bits - total tax of all marks (unpack, sekvences, endmark)
int sumsek[PACKTYPES];		// Number of sekvences in final pack model

int bufoff[MAXLEN][PACKTYPES];	// Offset of   packed sequecne ended in this byte
int buflen[MAXLEN][PACKTYPES];	// Length of   packed sequecne ended in this byte
int buftax[MAXLEN][PACKTYPES];	// Total price of packed data ending in this byte

///////////////////////////
// Types of compressions //
///////////////////////////

// Systems for coding length and unpacked data:

#define TYPLZM 1	// Standart LZM compression - 7 bit length for both - sekvence and unpacked block
#define TYPLZE 2	// Standart LZE compression - 6/14 bit length for both - sekvence and unpacked block
#define TYPZX7 3	// Original ZX7 compression - 1 bit for each unpacked byte and 1 bit + elias-gama length for sekvence
#define TYPBLK 4	// Bitstream BLK first simple - elias-gama length for both sekvence and unpacked block + 1 bit for sekvence / unpacked
#define TYPBS1 5	// Bitstream LZX standart 1 - 0=unpack byte, 10=2 byte sek, 110=3 byte sek, 1110 + EG length = sekvence, 1111 + EG length = unpacked block

// Systems for coding offsets:

#define POSLZM 1	// Standart for LZM compression - one byte (8 bit) offset
#define POSLZE 2	// Standart for LZE compression - one or two byte (7 or 15 bit) offset
#define POSOF4 4	// Four step offsets (1,2,3,4 or 2,4,6,8 or 3,6,9,12 or 4,8,12,16 bits)
#define POSOF1 5	// One fixed offset used for all sekvences
#define POSOF2 6	// Two fixed offsets (similar than ZX7)
#define POSOFD 7	// Simple elias-gama coding offset

void DisplayCompression(int comsel)
{
	int com = typcom[comsel];
	int off = typpos[comsel];

	switch (com)
	{
	case TYPLZM: printf("LZM"); break;
	case TYPLZE: printf("LZE"); break;
	case TYPZX7: printf("ZX7"); break;
	case TYPBLK: printf("BLK"); break;
	case TYPBS1: printf("BS1"); break;
	default: printf("???");
	}

	putchar(' ');
	int z = 0;

	switch (off)
	{
	case POSLZM: z = printf("offset 1 byte"); break;
	case POSLZE: z = printf("offset 2 bytes"); break;
	case POSOF4: z = printf("4 offs: %u %u %u %u", ofset1[comsel], 2 * ofset1[comsel], 3 * ofset1[comsel], 4 * ofset1[comsel]); break;
	case POSOF1: z = printf("fixed offset %2u", ofset1[comsel]); break;
	case POSOF2: z = printf("two offsets %2u %u", ofset1[comsel], ofset2[comsel]); break;
	case POSOFD: z = printf("elias-gama offset"); break;
	default:	 z = printf("(unknown)");
	}
	printf(" %s ", ".................." + z);
	z = printf("t%c%c", com + '0', off + '0');
	switch (off)
	{
	case POSOF4:
	case POSOF1: z += printf("o%u", ofset1[comsel]); break;
	case POSOF2: z += printf("o%uo%u", ofset1[comsel], ofset2[comsel]); break;
	}

	printf(" %s ", ".........." + z);
}

int GetBitsNumber(int number)
{
	if (number < 1) return 0;
	if (number < 2) return 1;
	if (number < 4) return 2;
	if (number < 8) return 3;
	if (number < 16) return 4;
	if (number < 32) return 5;
	if (number < 64) return 6;
	if (number < 128) return 7;
	if (number < 256) return 8;
	if (number < 512) return 9;
	if (number < 1024) return 10;
	if (number < 2048) return 11;
	if (number < 4096) return 12;
	if (number < 8192) return 13;
	if (number < 16384) return 14;
	if (number < 32768) return 15;
	if (number < 65536) return 16;
	return 999999;
}

int StandaloneFirstByte(int com)
{
	switch (com)
	{
	case TYPLZM:
	case TYPLZE:
	case TYPBLK: return 0;
	case TYPZX7:
	case TYPBS1: return 1;
	default: return -1;
	}
}

int GetNumberOffsets(int off)
{
	switch (off)
	{
	case POSOF4:
	case POSOF1: return 1;
	case POSOF2: return 2;
	case POSLZM:
	case POSLZE:
	case POSOFD: return 0;
	default: return -1;
	}
}

int GetTaxUnpack(int comsel, int length)
{
	switch (typcom[comsel])
	{
	case TYPLZM: return (length + 126) / 127 << 3;
	case TYPLZE: return length < 64 ? 8 : length < 16384 ? 16 : length < 16384 + 63 ? 24 : 32;
	case TYPZX7: return length;
	case TYPBLK: return 2 * GetBitsNumber(length);
	case TYPBS1: return length < 8 ? length : 2 * GetBitsNumber(length);
	}
	return 0;
}

int GetBitsEndMark(int comsel)
{
	switch (typcom[comsel])
	{
	case TYPLZM:
	case TYPLZE: return 8;
	case TYPZX7:
	case TYPBLK:
	case TYPBS1: return 18;
	}
	return 0;
}

void SetLimitArray(int comsel)
{
	static int make_error[] = { 0, 0 };

	static int lzm_off_lim[] = { 255, 0 };
	static int lzm_len_lim[] = { 127, 0 };
	static int lzm_all_tax[] = { 8, 0 };

	static int lze_off_lim[] = { 127, 32767, 0 };
	static int lze_len_lim[] = { 63, 16383, 0 };
	static int lze_all_tax[] = { 8, 16, 0 };

	static int blk_len_lim[] = { 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 0};
	static int blk_len_tax[] = { 2, 4, 6,  8, 10, 12,  14,  16,  18,   20,   22,   24,   26,    28,    30,    32, 0};

	static int blk_off_lim[] = { 1, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static int blk_off_tax[] = { 1, 3, 5,  7,  9, 11,  13,  15,  17,   19,   21,   23,   25,    27,    29,    31, 0 };

	static int lzx_len_lim[] = { 2, 3, 7, 15, 31, 63, 127, 255, 511, 1023, 2047, 4095, 8191, 16383, 32767, 65535, 0 };
	static int lzx_len_tax[] = { 2, 3, 7,  9, 11, 13,  15,  17,  19,   21,   23,   25,   27,    29,    31,    33, 0 };

	int *lo = make_error;
	int *to = make_error;
	int *ll = make_error;
	int *tl = make_error;

	switch (typcom[comsel])
	{
	case TYPLZM: ll = lzm_len_lim; tl = lzm_all_tax; break;
	case TYPLZE: ll = lze_len_lim; tl = lze_all_tax; break;
	case TYPZX7:
	case TYPBLK: ll = blk_len_lim; tl = blk_len_tax; break;
	case TYPBS1: ll = lzx_len_lim; tl = lzx_len_tax; break;
	}

	int o1 = ofset1[comsel];
	int o2 = ofset2[comsel];

	switch (typpos[comsel])
	{
	case POSLZM: lo = lzm_off_lim; to = lzm_all_tax; break;
	case POSLZE: lo = lze_off_lim; to = lze_all_tax; break;
	case POSOFD: lo = blk_off_lim; to = blk_off_tax; break;

	case POSOF1:
		lo = tstalloc(4);
		lo[0] = 1 << o1;
		lo[1] = 0;
		to = lo + 2;
		to[0] = o1;
		to[1] = 0;
		break;

	case POSOF2:
		lo = tstalloc(6);
		lo[0] = 1 << o1;
		lo[1] = (1 << o1) + (1 << o2);
		lo[2] = 0;
		to = lo + 3;
		to[0] = o1 + 1;
		to[1] = o2 + 1;
		to[2] = 0;
		break;

	case POSOF4:
		int offs = 0;
		lo = tstalloc(10); to = lo + 5;
		for (int i = 0; i < 4; i++)
		{
			int bits = ((i + 1) * o1);
			offs += 1 << bits;
			lo[i] = offs;
			to[i] = bits + 2;
		}
		lo[4] = 0;
		to[4] = 0;
		break;
	}

	limoff[comsel] = lo;
	taxoff[comsel] = to;
	limlen[comsel] = ll;
	taxlen[comsel] = tl;
}

int SetUnpackTax(int comsel)
{
	loccnt[comsel]++;
	int cnt = loccnt[comsel];
	switch (typcom[comsel])
	{
	case TYPLZM:
		if (cnt == 128) { loccnt[comsel] = 1; return 16; }
		if (cnt == 1) return 16;
		return 8;
	case TYPLZE:
		if (cnt == 16384) { loccnt[comsel] = 1; return 16; }
		if (cnt == 64) return 16;
		if (cnt == 1) return 16;
		return 8;
	case TYPZX7:
		return 9;
	case TYPBLK:
		if (cnt < 3 ||
			cnt == 4 ||
			cnt == 8 ||
			cnt == 16 ||
			cnt == 32 ||
			cnt == 64 ||
			cnt == 128 ||
			cnt == 256 ||
			cnt == 512 ||
			cnt == 1024 ||
			cnt == 2048 ||
			cnt == 4096 ||
			cnt == 8192 ||
			cnt == 16384 ||
			cnt == 32768) return 10;
		return 8;
	case TYPBS1:
		if (cnt <= 8) return 9;
		if (cnt == 16 ||
			cnt == 32 ||
			cnt == 64 ||
			cnt == 128 ||
			cnt == 256 ||
			cnt == 512 ||
			cnt == 1024 ||
			cnt == 2048 ||
			cnt == 4096 ||
			cnt == 8192 ||
			cnt == 16384 ||
			cnt == 32768) return 10;
		return 8;
	}
	return 1000000;
}

void PutBit(int bit)
{
	if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" %c", bit ? '1':'0');
	if (!bits_stream)
	{
		poke_stream = packed_index;
		packed_index++;
		bits_stream++;
	}
	bits_stream = (bits_stream << 1) | (bit ? 1 : 0);
	if (bits_stream & 0xFF00)
	{
		packed_data[poke_stream] = (unsigned char)(bits_stream & 0xFF);
		bits_stream = 0;
	}
	overhead_bits++;
}

void PutValue(int value, int bits)
{
	if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" (Val:%u B:%u)", value, bits);
	if (bits) for (int mask = 1 << (bits - 1); mask; mask >>= 1) PutBit(value & mask);
}

void PutEliasGama(int value, int bits)
{
	if (!value) { printf("\nError in PutEliasGama: Zero input value. Bits:%u\n", bits); exit(7); }
	if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" (Eli:%u B:%u)", value, bits);
	while (value > (1 << (bits + 1)) - 1) { bits++; PutBit(0); }
	PutBit(1);
	PutValue(value, bits);
}

void PutEndMark(int comsel)
{
	int bits = 16 + 1;
	switch (typcom[comsel])
	{
	case TYPLZM:
	case TYPLZE:
		packed_data[packed_index++] = 0;
		overhead_bits += 8;
		if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" 0x00");
		return;
	case TYPBS1:
		bits = 13;
		PutBit(1);
		PutBit(1);
		PutBit(1);
		PutBit(1);
	// case TYPZX7:
	// case TYPBLK:
	}
	for (int i = 0; i < bits; i++) PutBit(0);
	PutBit(1);
}

void PutSekvence(int comsel, int offset, int length)
{
	if (offset > maxoff[comsel] || length > maxlen[comsel] || length < 2)
	{
		printf("\nError in PutSekvence: Offset %u > %u or length %u > %u in ", offset, maxoff[comsel], length, maxlen[comsel]);
		DisplayCompression(comsel);
		putchar('\n');
		exit(6);
	}

	switch (typcom[comsel])
	{
	case TYPLZM:
		packed_data[packed_index++] = (length << 1) | 0x01;
		if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" %02X", (length << 1) | 0x01);
		overhead_bits += 8;
		break;

	case TYPLZE:
		if (length < 64)
		{
			packed_data[packed_index++] = length | 0x40;
			if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" %02X", length | 0x40);
			overhead_bits += 8;
		}
		else
		{
			packed_data[packed_index++] = (length >> 8) | 0xC0;
			packed_data[packed_index++] = length & 0xFF;
			if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf("  %02X %02X", (length >> 8) | 0xC0, length & 0xFF);
			overhead_bits += 16;
		}
		break;

	case TYPBLK:
		PutEliasGama(length - 1, 0);
		if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" (pack)");
		PutBit(0);
		break;

	case TYPZX7:
		PutBit(0);
		PutEliasGama(length - 1, 0);
		break;

	case TYPBS1:
		if (length == 2)
		{
			PutBit(1);
			PutBit(0);
		}
		else if (length == 3)
		{
			PutBit(1);
			PutBit(1);
			PutBit(0);
		}
		else
		{
			PutBit(1);
			PutBit(1);
			PutBit(1);
			PutBit(0);
			PutEliasGama(length, 2);
		}
		break;
	}

	if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" -");

	int i;
	int o1 = ofset1[comsel];

	switch (typpos[comsel])
	{
	case POSLZM:
		packed_data[packed_index++] = offset;
		if (ENAOUT) printf(" %02X", offset);
		overhead_bits += 8;
		break;

	case POSLZE:
		if (offset < 128)
		{
			packed_data[packed_index++] = offset;
			if (ENAOUT) printf(" %02X", offset);
			overhead_bits += 8;
		}
		else
		{
			packed_data[packed_index++] = (offset >> 8) | 0x80;
			packed_data[packed_index++] = offset & 0xFF;
			if (ENAOUT) printf(" %02X %02X", (offset >> 8) | 0x80, offset & 0xFF);
			overhead_bits += 16;
		}
		break;

	case POSOFD:
		PutEliasGama(offset, 0);
		break;

	case POSOF1:
		PutValue(offset - 1, o1);
		break;
			
	case POSOF2:
		if (offset <= (1 << o1))
		{
			PutBit(1);
			PutValue(offset - 1, o1);
		}
		else
		{
			PutBit(0);
			PutValue(offset - (1 << o1) - 1, ofset2[comsel]);
		}
		break;

	case POSOF4:
		for (i = 0; i < 4; i++) if (offset <= limoff[comsel][i] || !limoff[comsel][i]) break;
		if (i) offset -= limoff[comsel][i - 1];
		PutBit(i & 0x02);
		PutBit(i & 0x01);
		PutValue(offset - 1, o1*(i + 1));
		break;
	}
}

void PutUnpacked(int comsel, int index, int length)
{
	switch (typcom[comsel])
	{
	case TYPLZM:
		while (length)
		{
			int len = length > 0x7F ? 0x7F : length;
			packed_data[packed_index++] = len << 1;
			if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" %02X", len);
			memcpy(packed_data + packed_index, input_data + index, len);
			packed_index += len;
			overhead_bits += 8;
			length -= len;
			index += len;
		}
		break;

	case TYPLZE:
		while (length)
		{
			int len = length > 0x3FFF ? 0x3FFF : length;
			if (len < 0x40)
			{
				packed_data[packed_index++] = len;
				if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" %02X ", len);
				overhead_bits += 8;
			}
			else
			{
				packed_data[packed_index++] = (len >> 8) | 0x80;
				packed_data[packed_index++] = len & 0xFF;
				if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" %02X %02X ", (length >> 8) | 0x80, len & 0xFF);
				overhead_bits += 16;
			}
			memcpy(packed_data + packed_index, input_data + index, length);
			packed_index += len;
			length -= len;
			index += len;
		}
		break;

	case TYPBLK:
		PutEliasGama(length, 0);
		if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" (unpack)");
		PutBit(1);
		memcpy(packed_data + packed_index, input_data + index, length);
		packed_index += length;
		break;

	case TYPZX7:
		for (int i = 0; i < length; i++)
		{
			PutBit(1);
			packed_data[packed_index++] = input_data[index + i];
		}
		break;

	case TYPBS1:
		if (length < 8)
		{
			for (int i = 0; i < length; i++)
			{
				PutBit(0);
				packed_data[packed_index++] = input_data[index + i];
			}
		}
		else
		{
			for (int i = 0; i < 4; i++) PutBit(1);
			if (ENAOUT && index >= ENAZAC && index <= ENAEND) printf(" -");
			PutEliasGama(length, 3);
			memcpy(packed_data + packed_index, input_data + index, length);
			packed_index += length;
		}
		break;
	}
}

void SetCompression(int compress_type, int compress_offset, int off1, int off2)
{
	int i;
	if (compressions >= PACKTYPES) {compressions++;return;}
	if (StandaloneFirstByte(compress_type) < 0 || GetNumberOffsets(compress_offset) < 0) return;

	typcom[compressions] = compress_type;
	typpos[compressions] = compress_offset;
	ofset1[compressions] = off1;
	ofset2[compressions] = off2;
	SetLimitArray(compressions);
	for (i = 1; limoff[compressions][i]; i++); maxoff[compressions] = limoff[compressions][i - 1];
	for (i = 1; limlen[compressions][i]; i++); maxlen[compressions] = limlen[compressions][i - 1];

	declen[compressions] = 0;
	if (depackers)
	{
		for (int i = 0; i < depackers; i++)
			if (compress_type   == depacker_com[i]
			 && compress_offset == depacker_pos[i])
			{
				declen[compressions] = depacker_len[i];
				break;
			}
	}

	if (ENAINF)
	{
		printf("%5u ... ", compressions);
		DisplayCompression(compressions);
		printf("Max length: %-5u  Max offset: %-5u", maxlen[compressions], maxoff[compressions]);
		if (declen[compressions]) printf("    Depacker length: %u", declen[compressions]);
		putchar('\n');
		if (ENAINF > 1)
		{
			printf("\t\tOffset limits: "); for (int i = 0; limoff[compressions][i]; i++) printf("  %u.%u", limoff[compressions][i], taxoff[compressions][i]); putchar('\n');
			printf("\t\tLength limits: "); for (int i = 0; limlen[compressions][i]; i++) printf("  %u.%u", limlen[compressions][i], taxlen[compressions][i]); putchar('\n');
			putchar('\n');
		}
	}
	compressions++;
}

void PrepareAllCompressionTypes(int input_size)
{
	int maxbits = GetBitsNumber(input_size);
	int minbits = (maxbits >> 2) + 1;
	compressions = 0;

	if (ENAINF) printf("Used compression types   Length: %u  Bits: %u..%u  Com: -t%c%co%uo%u\n\n",
		input_size, minbits, maxbits, compress_type + '0', compress_posi + '0', offset1bits, offset2bits);

	for (int cmp = compress_type ? compress_type : 1; cmp <= (compress_type ? compress_type : 9); cmp++)
	{
		if (compress_type && compress_posi || (cmp >= TYPLZM && cmp <= TYPLZE))
		{
			if ((!compress_posi || compress_posi == POSLZM) && (compress_type || compress_posi || cmp == TYPLZM)) SetCompression(cmp, POSLZM, 0, 0);
			if ((!compress_posi || compress_posi == POSLZE) && (compress_type || compress_posi || cmp == TYPLZE)) SetCompression(cmp, POSLZE, 0, 0);
		}

		if (compress_type && compress_posi || (cmp >= TYPZX7 && cmp <= TYPBS1))
		{
			if (!compress_posi || compress_posi == POSOFD) SetCompression(cmp, POSOFD, 0, 0);

			if (!compress_posi || compress_posi == POSOF4)
			{
				if (offset1bits > 4) { printf("Error: Offset %u overflows maximal value 4\n", offset1bits); exit(1); }
				for (int i = offset1bits ? offset1bits : 1; i <= (offset1bits ? offset1bits : (maxbits + 1) >> 2); i++)
					SetCompression(cmp, POSOF4, i, 0);
			}

			if (!compress_posi || compress_posi == POSOF1)
				for (int i = offset1bits ? offset1bits : minbits; i <= (offset1bits ? offset1bits : maxbits); i++)
					SetCompression(cmp, POSOF1, i, 0);

			if (!compress_posi || compress_posi == POSOF2)
			{
				if (offset1bits && !offset2bits) { puts("Error: 2nd offset is not set."); exit(1); }
				for (int i = offset2bits ? offset2bits : minbits; i <= (offset2bits ? offset2bits : maxbits); i++)
					for (int j = offset1bits ? offset1bits : 1; j <= (offset1bits ? offset1bits : i - 2); j++)
						SetCompression(cmp, POSOF2, j, i);
			}
		}
	}

	if (ENAINF) putchar('\n');
	if (!compressions) { printf("\nError: Not valid compression found.\n"); exit(1); }
	if (compressions > PACKTYPES) { printf("\nError: Number of compression types %u overflows %u.\n", compressions, PACKTYPES); exit(1); }
}

int GetTaxSekvence(int comsel, int offset, int length)
{
	if (offset < 1 || length < 2) { printf("ERROR: GetTaxSekvence:  Offset = %d  Length = %d\n", offset, length); exit(1); }

	int o, * offsets = limoff[comsel]; for (o = 0; offsets[o]; o++) if (offset <= offsets[o]) break;
	int l, * lengths = limlen[comsel]; for (l = 0; lengths[l]; l++) if (length <= lengths[l]) break;
	return (taxoff[comsel][o] + taxlen[comsel][l]);
}

int * tstalloc(int num)
{
	int * memory = (int*)malloc(num * sizeof(int));
	if (memory) return memory;
	printf("ERROR: malloc(%u) returns null\n", num * sizeof(int));
	exit(1); 
}

void DisplayPackModel(int comsel, int input_size)
{
	if (ENADSP)
	{
		printf("\nPack model for ");
		DisplayCompression(comsel);
		puts("\n");
	}
	int index = 0;
	int numsek = 0;
	int locpck = 0;
	int locunp = 0;
	int locove = 0;
	int totpck = 0;
	int totunp = 0;
	int totove = GetBitsEndMark(comsel);

	if (StandaloneFirstByte(typcom[comsel]))
	{
		index = 1;
		totunp = 1;
		if (ENADSP && !ENAZAC) printf("    +0 0000 - 0000 0      First byte 0x%02X\n", *input_data);
	}
	while(index < input_size)
	{
		int length = buflen[index][comsel];
		int offset = bufoff[index][comsel];
		if (offset)
		{
			locove = GetTaxSekvence(comsel, offset, length);

			if (ENADSP && index + length >= ENAZAC && index <= ENAEND)
				printf("%+6d %04X - %04X %-5u  Tax:%-3u  Len:%-5u Sekvence offset:%-5u \n",
					index, index, index + length - 1, index + length - 1, locove, length, offset);

			totove += locove;
			totpck += length;
			index += length;
			numsek++;

			if (length != buflen[index - 1][comsel] || offset != bufoff[index - 1][comsel])
				printf("\nError in pack model: Sekvence at %+d  Len: %u<>%u Off: %u<>%u\n",
					index - length, length, buflen[index - 1][comsel], offset, bufoff[index - 1][comsel]);
		}
		else
		{
			locove = GetTaxUnpack(comsel, length);

			if (ENADSP && index + length >= ENAZAC && index <= ENAEND)
			{
				int i;
				printf("%+6d %04X - %04X %-5u  Tax:%-3u  Len:%-5u Data:",
					index, index, index + length - 1, index + length - 1, locove, length);
				for (i = 0; i < 16 && i < length; i++) printf(" %02X", input_data[index + i]);
				if (i < length)printf(" ...");
				putchar('\n');
			}

			totove += locove;
			totunp += length;
			index += length;

			if (length != buflen[index - 1][comsel] || offset || bufoff[index - 1][comsel])
				printf("\nError in pack model: Unpacked at: %+d  Len: %u<>%u Off: %u %u\n",
					index - length, length, buflen[index - 1][comsel], offset, bufoff[index - 1][comsel]);
		}
	}
	if (ENADSP) putchar('\n');
	if (ENADSP)
	{
		DisplayCompression(comsel);
		printf(" %5u => %-5u    Saved data: %-5u  Unpack data: %-5u  Overhead: %6u bits  %5u bytes\n",
			input_size, totunp + ((totove + 7) >> 3), totpck, totunp, totove, (totove + 7) >> 3);
	}
	if (ENADSP) puts("\n");
	
	sumsek[comsel] = numsek;
	sumpck[comsel] = totpck;
	sumunp[comsel] = totunp;
	sumove[comsel] = totove;
	sumall[comsel] = (totunp << 3) + totove;
}

int OptimizeSekvence(int comsel, int datzac, int sekzac, int sekend, int datend)
{
	if (!buflen[sekzac][comsel] || !buflen[sekend - 1][comsel]) printf("\nError in OptimizeSekvence: No sekvence. %u-%u-%u-%u\n", datzac, sekzac, sekend, datend);
	int lenzac = sekzac - datzac;
	int lenend = datend - sekend;
	int lensek = sekend - sekzac;
	int lenall = datend - datzac;
	int numsek = 0;
	int taxzac = 0; if (lenzac) taxzac = GetTaxUnpack(comsel, lenzac);
	int taxend = 0; if (lenend) taxend = GetTaxUnpack(comsel, lenend);
	int taxsek = 0; for (int idx = sekzac; idx < sekend; idx += buflen[idx][comsel], numsek++) taxsek += GetTaxSekvence(comsel, bufoff[idx][comsel], buflen[idx][comsel]);
	int taxnew = (lensek << 3) + GetTaxUnpack(comsel, lenall);
	int taxold = taxzac + taxsek + taxend;

	if (ENAOPT && datend >= ENAZAC && datzac <= ENAEND)
	{
		if (lenzac)
			printf("Test Unp %5u..%-5u Len:%-5u Tax:%-5u + ", datzac, sekzac - 1, lenzac, taxzac);
		else
			printf("Test  ....................................  ");

		printf("Sek %5u..%-5u Len:%-5u Num:%-5u Tax:%-5u", sekzac, sekend - 1, lensek, numsek, taxsek);

		if (lenend)
			printf(" + Unp %5u..%-5u Len:%-5u Tax:%-5u", sekend, datend - 1, lenend, taxend);
		else
			printf("  ...................................  ");

		printf(" %5u => %-5u Unp %5u..%-5u Len:%-5u Tax:%-5u ",
			taxold, taxnew,	datzac, datend - 1, lenall, GetTaxUnpack(comsel, lenall));
	}
	
	if (taxnew > taxold)
	{
		if (ENAOPT && datend >= ENAZAC && datzac <= ENAEND) printf("... (none)\n");
		return 0;
	}

	bufoff[datzac][comsel] = 0;
	buflen[datzac][comsel] = lenall;
	bufoff[datend - 1][comsel] = 0;
	buflen[datend - 1][comsel] = lenall;

	if (ENAOPT && datend >= ENAZAC && datzac <= ENAEND)
	{
		printf("... Removed");
		if (taxold - taxnew) printf(" %+d", taxold-taxnew);
		putchar('\n');
	}
	return 1;
}

int main(int number_of_params, char **parameters)
{
	puts("\nBusy soft: Universal LZX Packer 01");

	if (number_of_params < 2)
	{
		puts(
			"\nUse: LzxPack filename <options>\n\n"
			"Options:\n"
			" -s ......... show statistic of all used compressions  (default: not show)\n"
			" -a ......... save all compressed files  (default: save only the best result)\n"
			" -dfile ..... txt file with depacker lengths to assume them into statistic\n"
			" -tXYoAoB ... select required compression  (default: trying any compression)\n\n"
			"Selecting compression:\n"
			"  X ... compression type: 0=any 1=LZM 2=LZE 3=ZX7 4=BLK 5=BS1\n"
			"  Y ... coding offset: 0=any 1=LZM 2=LZE 4=OF4 5=OF1 6=OF2 7=OFD\n"
			"  A ... number of bits of offset for OF4 OF1 OF2\n"
			"  B ... number of bits of second offset for OF2 only\n"
			"Not presenting params A and B (only -tXY) or zero assumes any possible values.\n"
		);
		exit(1);
	}

	putchar('\n');

	//////////////////////////////////
	// Read command line parameters //
	//////////////////////////////////

	int make_all_compressions = 0;
	int show_statistics = 0;
	int name_indent = 0;
	char *name_inp = NULL;
	char *name_dec = NULL;

	compress_type = 0;
	compress_posi = 0;
	offset1bits = 0;
	offset2bits = 0;

	depackers = 0;

	for (int i = 1; i < number_of_params; i++)
	{
		if (parameters[i][0] == '-')
		{
			char * xx = parameters[i] + 1;
			if      (*xx == 'A' || *xx == 'a') make_all_compressions = 1;
			else if (*xx == 'S' || *xx == 's') show_statistics = 1;
			else if (*xx == 'T' || *xx == 't')
			{
				xx++;
				if (!*xx) { puts("Error: Compression type -tXX not selected"); exit(1); }
				if (*xx == '0' || StandaloneFirstByte(*xx - '0') >= 0) compress_type = *xx - '0';
				else { printf("Error: Unknown compression type '%c'\n", *xx); exit(1); }
				xx++;
				if (!*xx) { puts("Error: Coding offset -tXX not selected"); exit(1); }
				if (*xx == '0' || GetNumberOffsets(*xx - '0') >= 0) compress_posi = *xx - '0';
				else { printf("Error: Unknown coding offset '%c'\n", *xx); exit(1); }
				xx++;
				if (*xx == 'O' || *xx == 'o')
				{
					while (*xx)
					{
						xx++;
						if (*xx <'0' || *xx > '9') break;
						offset1bits = 10 * offset1bits + *xx - '0';
					}

					if (*xx == 'O' || *xx == 'o')
					{
						while (*xx)
						{
							xx++;
							if (*xx <'0' || *xx > '9') break;
							offset2bits = 10 * offset2bits + *xx - '0';
						}
					}
				}
				if (*xx) { printf("Error: Unexpected characters in type selection: %s\n", xx); exit(1); }

				if (offset1bits > 16 || offset2bits > 16) { puts("Error: Offset overflows maximal value 16"); exit(1); }

				if (offset2bits && (offset2bits <= offset1bits || offset2bits <= 2))
				{
					printf("Error: 2nd offset %u must be bigger than %s\n", offset2bits, offset2bits <= 2 ? "2" : "1st offset"); exit(1);
				}
			}
			else if (*xx == 'D' || *xx == 'd')
			{
				xx++;
				if (!strlen(xx)) { puts("Error: No filename in -d option"); exit(1); }

				char line[1024];
				name_dec = xx;
				depackers = 0;
				memset(line, 0x00, 1024);
				FILE *file_inp = fopen(xx, "rt"); if (!file_inp) { perror(xx); exit(1); }
				while (fgets(line, sizeof(line), file_inp))
				{
					line[1024] = 0;
					if (StandaloneFirstByte(line[0] - '0') >= 0 && GetNumberOffsets(line[1] - '0') >= 0)
					{
						if (depackers < DEPACKERS)
						{
							depacker_com[depackers] = line[0] - '0';
							depacker_pos[depackers] = line[1] - '0';
							depacker_len[depackers] = atoi(line + 3);
						}
						depackers++;
					}
					line[3] = 0;
				}
				fclose(file_inp);
				if (!depackers) { printf("\nError: Not valid depacked length definition found in file %s\n", xx); exit(1); }
				if (depackers > DEPACKERS) { printf("\nError: Number of depackers %u overflows %u.\n", depackers, DEPACKERS); exit(1); }
			}
			else { printf("Error: Unknown option %s\n", parameters[i]); exit(1); }
		}
		else
		{
			if (name_inp) { printf("Error: Too many input files:  %s  %s\n", name_inp, parameters[i]); exit(1); }
			name_inp = parameters[i];
		}
	}
	
	if (!name_inp) { puts("Error: Missing input file !"); exit(1); }

	if (compress_type || compress_posi || offset1bits || offset2bits)
		printf("Selected compression: -t%c%co%uo%u\n\n",compress_type+'0', compress_posi+'0', offset1bits, offset2bits);

	if (depackers && show_statistics)
	{
		printf("%u depackers found in file %s\n\n", depackers, name_dec);
		for (int i = 0; i < depackers; i++)
			printf("Compress: %c%c ... depacker length: %u\n",
				depacker_com[i] + '0', depacker_pos[i] + '0', depacker_len[i]);
		putchar('\n');
	}

	/////////////////////
	// Read input file //
	/////////////////////

	FILE *file_inp = fopen(name_inp,"rb"); if (!file_inp) {perror(name_inp);exit(1);}
	int input_size = fread(input_data, 1, MAXLEN - 1, file_inp);
	fclose(file_inp);

	strcpy(name_out, name_inp);
	int name_length = strlen(name_out);
	if (name_length > 5)
	{
		char * dotpos = strrchr(name_out + name_length - 5, '.');
		if (dotpos) { *dotpos = 0; name_length = dotpos - name_out; }
	}

	name_indent = name_length + 15;
	printf("Load file: %-*s  Length: %u\n", name_indent, name_inp, input_size);
	if (!input_size) {puts("Error: Zero length of input file !");exit(1);}

	////////////////////////////////////////
	// Prepare and check all compressions //
	////////////////////////////////////////

	PrepareAllCompressionTypes(input_size);

	if (ENAINF > 1)	// Display taxes and limits for offsets and lengths of sekvences //
	{
		#define VALS 42
		int values[VALS];
		for (int idx = 0, val = 1; idx < VALS; idx++, val += val < 10 ? 1 : val >> 2) values[idx] = val;
		for (int s = 0; s < 3; s++)
		{
			printf("Test for limits ....... %s", !s ? "Offsets ... " : s==1 ? "Pack length " : "Unpack len  ");
			for (int idx = 0; idx < VALS; idx++) printf(idx < 16 ? " %3u" : " %4u", values[idx]); printf("\n\n");
			for (int com = 0; com < compressions; com++)
			{
				DisplayCompression(com);
				if (s==1) printf("    ");
				int *limits = s ? limlen[com] : limoff[com];
				for (int idx = s==1 ? 1 : 0; idx < VALS; idx++)
				{
					int i, tax;
					if (s < 2)
					{
						for (i = 0; limits[i] && values[idx] > limits[i]; i++);
						tax = limits[i] ? (s ? taxlen[com][i] : taxoff[com][i]) : 0;
					}
					else
						tax = GetTaxUnpack(com, values[idx]);

					if (tax)
						printf(idx < 16 ? " %3u" : " %4u", tax);
					else
						printf(idx < 16 ? "   -" : "    -");
				}
				putchar('\n');
			}
			putchar('\n');
		}
	}

	if (ENAINF > 2) for (int com = 0; com < compressions; com++) // Test GetTaxSekvence() //
	{
		printf("Test GetTaxSekvence for: "); DisplayCompression(com); printf("\n\nLength =>");
		for (int len = 2; len < 300; len += len < 10 ? 1 : len >> 2) printf("%4u ", len);
		printf("\nOffset ,"); for (int i = 1; i < 126; i++) putchar('-'); putchar('\n');
		for (int off = 1; off < 300; off += off < 10 ? 1 : off >> 2)
		{
			printf("%5u  |", off);
			for (int len = 2; len < 300; len += len < 10 ? 1 : len >> 2)
				if (maxoff[com] < off || maxlen[com] < len)
					printf("    -");
				else
					printf(" %4u", GetTaxSekvence(com, off, len)); putchar('\n');
		}
		putchar('\n');
	}

	////////////////////////////////
	// Main LZ compression engine //
	////////////////////////////////

	clock_t run_begin = clock();

	int cmps;
	if (ENACOM) cmps = 0;
	if (ENACOM) printf("Searching sekvences ...\n\n");

	unsigned char byte = input_data[0];
	if (ENACOM > 1 && ENAZAC == 0) printf("    +0 0000  Byte: 0x%02X '%c'  First byte\n",byte,byte >= 32 && byte <= 126 ? byte : ' ');

	// Set tax for first byte from input file //
	memset(loccnt, 0x00, sizeof(loccnt));
	for (int comsel = 0; comsel < compressions; comsel++)
	{
		if (StandaloneFirstByte(typcom[comsel]))
		{
			buftax[0][comsel] = 8;
		}
		else
		{
			buftax[0][comsel] = SetUnpackTax(comsel);
		}
		bufoff[0][comsel] = 0;
		buflen[0][comsel] = 0;
	}

	// Initializations //
	int predat = 0;
	memset(preoff, 0x00, sizeof(preoff));
	memset(prelen, 0x00, sizeof(prelen));

	repper[0] = 0;
	replen[0] = 0;
	lstidx[0] = 0;
	lstlen[0] = 0;
	bufnxt[0] = 0;

	// Main loop for all next bytes from input file //
	for(index = 1; index < input_size; index++)
	{
		// Create word hash for quick searching //
		int hash_index = input_data[index] | (input_data[index-1] << 8);
		int locate = hash_table[hash_index];
		bufnxt[index] = locate;
		hash_table[hash_index] = index;

		if (ENACOM > 2 && index >= ENAZAC && index <= ENAEND)
		{
			int max = 8;
			unsigned char byte = input_data[index];
			printf("%+6d %04X  Byte: 0x%02X '%c'  Create hash: 0x%04X  ",
				index,index,byte,
				byte >= 32 && byte <= 126 ? byte : ' ',
				hash_index);
			int nxt = bufnxt[index];
			if (nxt) printf("  Next hashes at offsets:");
			while(nxt && max)
			{
				printf(" %u",nxt);
				nxt = bufnxt[nxt];
				max--;
			}
			if (nxt) printf(" ...");
			putchar('\n');
		}

		// Set initial unpack tax for new processed byte //
		for (int comsel = 0; comsel < compressions; comsel++)
		{
			bufoff[index][comsel] = 0;
			buflen[index][comsel] = 0;
			buftax[index][comsel] = buftax[index - 1][comsel] + SetUnpackTax(comsel);

			if (ENAUNP && index >= ENAZAC && index <= ENAEND)
			{
				printf("\t\t\t\t\t");
				DisplayCompression(comsel);
				int previous_tax = buftax[index - 1][comsel];
				int tax = buftax[index][comsel] - previous_tax;
				printf("  Cnt: %-5u  Unpack tax: %6u +8 ", loccnt[comsel], previous_tax);
				if (tax != 8) printf("%+d", tax - 8); else printf("  ");
				printf(" = %-6u  bits\n", previous_tax + tax);
			}
		}

		memset(locskp, 0x00, sizeof(locskp));
		memset(idxoff, 0x00, sizeof(idxoff));
		memset(idxlen, 0x00, sizeof(idxlen));

		int needed = 1;

		// Direct detection for next best sekvence for compressions //
		/* (not used because it has influence for compression ratio)
		int needed = 0;
		if (locate)
		{
			if (ENAPRV && index >= ENAZAC && index <= ENAEND) printf("\t\t\t\tTest previous sekvences\n");
			for (int comsel = 0; comsel < compressions; comsel++)
			{
				if (ENAPRV && index >= ENAZAC && index <= ENAEND) { printf("\t\t\t\t\t"); DisplayCompression(comsel); }

				int offset = bufoff[index - 1][comsel];

				if (ENACOM && offset) cmps++;
				if (offset && input_data[index] == input_data[index - offset])
				{
					int seklen = buflen[index - 1][comsel] + 1;

					if (ENAPRV && index >= ENAZAC && index <= ENAEND)
						printf("  Sek. at %+-5d Off:%-5u Len:%-5u  Idx:%-2u Lim:%-5u",
							index - 1, offset, seklen - 1, prelen[comsel], limlen[comsel][prelen[comsel]]);

					if (seklen <= limlen[comsel][prelen[comsel]])
					{
						locskp[comsel] = 1;
						loccnt[comsel] = 0;
						bufoff[index][comsel] = offset;
						buflen[index][comsel] = seklen;
						buftax[index][comsel] = buftax[index - 1][comsel];

						if (ENAPRV && index >= ENAZAC && index <= ENAEND)
							printf("  =>  Len:%-5u ... Accepted", seklen);
					}
					else
						if (ENAPRV && index >= ENAZAC && index <= ENAEND) printf(" (limit over)");
				}
				else
					if (ENAPRV && index >= ENAZAC && index <= ENAEND) printf(" (no previous sekvence)");

				if (!locskp[comsel])
				{
					needed = 1;
					preoff[comsel] = 0;
					prelen[comsel] = 0;
				}

				if (ENAPRV && index >= ENAZAC && index <= ENAEND) putchar('\n');
			}
		}
		else
		{
			if (predat)
			{
				predat = 0;
				memset(preoff, 0x00, sizeof(preoff));
				memset(prelen, 0x00, sizeof(prelen));
			}
		}
		*/

		int best_offset = 0;
		int best_length = 0;
		int repe_length = 0;
		int repe_period = 0;
		int maxloc = MAXLEN;

		// Loop for search all sekvences related to processed byte //
		while(locate && needed)
		{
			int offset = index - locate;

			if (locate < best_length)
			{
				if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
					printf("%+6d %04X                      "
						"Skip hash%+6d %04X   Offset:%-5d  "
						" and all next hashes due to bestlen = %u\n",
						index, index, locate, locate, -offset, best_length);
				break;
			}

			if (locate > maxloc)
			{
				if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
					printf("%+6d %04X                      "
						"Skip hash%+6d %04X   Offset:%-5d  "
						"due to maxloc = %u\n",
						index, index, locate, locate, -offset, maxloc);

				locate = bufnxt[locate];
				continue;
			}

			if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
				printf("%+6d %04X                      "
					"Test hash%+6d %04X   Offset:%-5d  "
					".....    RepPer:%-3u RepLen:%-5u    "
					"Previous is LstIdx:%-5u LstLen:%-5u \n",
					index, index, locate, locate, -offset,
					repper[locate], replen[locate],
					lstidx[locate - 1], lstlen[locate - 1]);

			int length = 0;
			///////////////////////////////////////////////////////
			if (lstidx[locate-1] == index-1)
			{
				length = lstlen[locate-1]+1;

				if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
					printf("%+6d %04X                        "
							"Previous sekvence at %+d  Len:%u  =>  Set length:%u\n",
							index, index, index-1, lstlen[locate-1], length);
			}
			///////////////////////////////////////////////////////
			else
			{
				int maxlen = locate + 1;
				int period = repper[locate];
				int subsub = index-period;
				if (period && period == repper[subsub])
				{
					if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
						printf("%+6d %04X                        "
								"Periodic sekvences detected:  "
								"%+d Period:%u Length:%u   and  "
								"%+d Period:%u Length:%u\n",
								index, index, locate, period, replen[locate],
								subsub,repper[subsub],replen[subsub]);

					int limit = period << 1; if (limit > maxlen) limit = maxlen;
					for (length = 2; length < limit; length++)
					{
						if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
							printf("%+6d %04X                          "
									"Compare before %+6d 0x%02X   %+6d 0x%02X   Len:%-5u \n",
									index,index,
									locate-length,input_data[locate-length],
									index -length,input_data[index -length],
									length);

						if (ENACOM) cmps++;
						if (input_data[locate-length] != input_data[index-length]) break;
					}
					if (length == limit)
					{
						length = period + replen[subsub];
						if (length > replen[locate]) length = replen[locate];
						if (length > locate + 1) length = locate + 1;
						maxloc = index - length + 1;
						if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
							printf("%+6d %04X                        "
									"Increase length %u to %u   Maxloc: %u\n",
									index,index,limit,length,maxloc);

						while(length < maxlen)
						{
							if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
								printf("%+6d %04X                          "
										"Compare after  %+6d 0x%02X   %+6d 0x%02X   Len:%-5u \n",
										index,index,
										locate-length,input_data[locate-length],
										index-length,input_data[index-length],
										length);

							if (ENACOM) cmps++;
							if (input_data[locate-length] != input_data[index-length]) break;
							length++;
						}
					}
				}

				else for (length = 2; length < maxlen; length++)
				{
					if (ENACOM > 4 && index >= ENAZAC && index <= ENAEND)
						printf("%+6d %04X                          "
							"Compare normal %+6d 0x%02X   %+6d 0x%02X   Len:%-5u\n",
							index, index,
							locate-length, input_data[locate-length],
							index -length, input_data[index -length],
							length);

					if (ENACOM) cmps++;
					if (input_data[locate-length] != input_data[index-length]) break;
				}
			}
			/////////////////////////
			lstidx[locate] = index;
			lstlen[locate] = length;
			/////////////////////////
			if (ENACOM > 2 && index >= ENAZAC && index <= ENAEND)
				printf("%+6d %04X                      Found sekvence.. "
					"Loc:%-+5d  Off:%-5d  Len:%-5u    RepPer:%-3u RepLen:%-5u\n",
					index, index, locate, -offset, length, repper[locate], replen[locate]);

			if (replen[locate])
			{
				int tmpper = repper[locate];
				int tmplen = replen[locate]; if (tmplen > length) tmplen = length;
				if (tmplen && (tmplen - tmpper > repe_length - repe_period) && (tmplen > tmpper))
				{
					repe_period = tmpper;
					repe_length = tmplen;
				}
			}

			if (best_length < length)
			{
				best_length = length;
				best_offset = offset;

				// Selecting and accepting sekvences for each compression //
				needed = 0;
				for (int comsel = 0; comsel < compressions; comsel++)
				{
					if (ENACPT && index >= ENAZAC && index <= ENAEND)
					{
						printf("\t\t\t\t\t");
						DisplayCompression(comsel);
						printf(" ... ");
					}
					if (locskp[comsel])
					{
						if (ENACPT && index >= ENAZAC && index <= ENAEND) printf("Skipped\n");
						continue;
					}
					if (offset > maxoff[comsel])
					{
						locskp[comsel] = 1;
						if (ENACPT && index >= ENAZAC && index <= ENAEND) printf("Skipped: Offset %u > MaxOff %u\n", offset, maxoff[comsel]);
						continue;
					}

					needed = 1;

					if (length < buflen[index][comsel])
					{
						if (ENACPT && index >= ENAZAC && index <= ENAEND) printf("Skipped: Length %u < BufLen %u\n", length, buflen[index][comsel]);
						continue;
					}

					// Find offset //
					int seloff = idxoff[comsel];
					if (offset > limoff[comsel][seloff])
					{
						while (limoff[comsel][seloff] && offset > limoff[comsel][seloff])
						{
							if (ENACPT && index >= ENAZAC && index <= ENAEND) printf(" %u", limoff[comsel][seloff]);
							seloff++;
						}

						if (!limoff[comsel][seloff])
						{
							locskp[comsel] = 1;
							if (ENACPT && index >= ENAZAC && index <= ENAEND) printf(" Skipped: Offset %u > LimOff %u\n", offset, limoff[comsel][seloff - 1]);
							continue;
						}

						idxoff[comsel] = seloff;
					}

					if (ENACPT && index >= ENAZAC && index <= ENAEND) printf(" o%u", limoff[comsel][seloff]);

					int bstlen = 0;
					int bstidx = 0;
					int bsttax = buftax[index][comsel];

					// Find and check length //
					int sellen = idxlen[comsel];
					while (limlen[comsel][sellen])
					{
						int cmptax = bsttax - taxoff[comsel][seloff] - taxlen[comsel][sellen];
						int tmplen = limlen[comsel][sellen];
						int tstend = tmplen; if (tstend > length) tstend = length;
						int tstzac = sellen ? (limlen[comsel][sellen - 1] + 1) : 2;
						if (tstzac < tstend - 8) tstzac = tstend - 8;

						if (ENACPT && index >= ENAZAC && index <= ENAEND) printf(" (%u)%u-%u", cmptax, tstzac, tstend);

						for (int tstlen = tstzac; tstlen <= tstend; tstlen++)
						{
							if (ENACPT && index >= ENAZAC && index <= ENAEND) printf(" %u(%u)", tstlen, buftax[index - tstlen][comsel]);

							if (buftax[index - tstlen][comsel] < cmptax)
							{
								cmptax = buftax[index - tstlen][comsel];
								bstlen = tstlen;
								bstidx = sellen;
								bsttax = cmptax + taxoff[comsel][seloff] + taxlen[comsel][bstidx];
								if (ENACPT && index >= ENAZAC && index <= ENAEND) putchar('*');
							}
						}
						if (ENACPT && index >= ENAZAC && index <= ENAEND) putchar(' ');

						if (length <= tmplen)
						{
							if (length == tmplen) sellen++;
							break;
						}

						sellen++;
					}
					idxlen[comsel] = sellen;
					
					if (bstlen)
					{
						predat = 1;
						preoff[comsel] = seloff;
						prelen[comsel] = bstidx;

						loccnt[comsel] = 0;
						bufoff[index][comsel] = offset;
						buflen[index][comsel] = bstlen;
						buftax[index][comsel] = bsttax;

						if (ENACPT && index >= ENAZAC && index <= ENAEND)
							printf("... Accept So:%u Sl:%u Len:%u Tax:%u)",
								seloff, bstidx, buflen[index][comsel], buftax[index][comsel]);

						if (bstlen == maxlen[comsel])
						{
							locskp[comsel] = 1;
							if (ENACPT && index >= ENAZAC && index <= ENAEND)
								printf(" ... Length %u == max length %u => skip all next matches", offset, maxoff[comsel]);
						}
					}
					if (ENACPT && index >= ENAZAC && index <= ENAEND) putchar('\n');

				}
				/////////////////////////////////////////////////////////////
			}

			if (ENACOM > 3 && !needed && index >= ENAZAC && index <= ENAEND)
				printf("%+6d %04X                      Skip all next hashes (not needed to find next sekvences)\n", index, index);

			locate = bufnxt[locate];
		}
		if (best_length)
		{
			if (best_length > best_offset)
			{
				repper[index] = best_offset;
				replen[index] = best_length + best_offset;
			}
			else
			{
				if (repe_length > best_length) repe_length = best_length;
				if (repe_length <= repe_period) {repe_length=0; repe_period =0;}
				repper[index] = repe_period;
				replen[index] = repe_length;
			}

			if (ENACOM > 1 && index >= ENAZAC && index <= ENAEND)
			{
				unsigned char byte = input_data[index];
				printf("%+6d %04X  ",index, index);
				if (ENACOM == 2)
					printf("Byte: 0x%02X '%c'",byte,byte >= 32 && byte <= 126 ? byte : ' ');
				else
					printf("..............");
				printf("    Best sekvence..... "
					"Loc:%-+5d  Off:%-5d  Len:%-5u    RepPer:%-3u RepLen:%-5u   Cmps:%u\n",
					index - best_offset, -best_offset, best_length,
					repper[index], replen[index],cmps);
			}
		}

		if (ENASUM && index >= ENAZAC && index <= ENAEND)
		{
			printf("\t\t\t\tSummmary of best sekvences\n");
			for (int comsel = 0; comsel < compressions; comsel++)
			{
				printf("\t\t\t\t\t");
				DisplayCompression(comsel);
				printf("  Index: %-5u  TotalTax: %-5u  ", index, buftax[index][comsel]);
				if (buflen[index][comsel])
					printf("Offset: %-5u  Length: %-5u\n", bufoff[index][comsel], buflen[index][comsel]);
				else
					puts("none");
			}
		}
		if (SEKNUM >= 0)
			printf(SEKNAM "  Index:%-5u  Offset:%-5u  Length:%-5u  Tax: %-5u  Dif: %d\n",
				index, bufoff[index][SEKNUM], buflen[index][SEKNUM],
				buftax[index][SEKNUM], buftax[index][SEKNUM] - buftax[index - 1][SEKNUM]);
	}
	// End of main LZ compression engine //
	if (ENACOM > 1 && index >= ENAZAC && index <= ENAEND) printf("%+6d %04X  End of file\n\n",input_size,input_size);

	/////////////////////////
	// Unreverse sekvences //
	/////////////////////////

	for (int comsel = 0; comsel < compressions; comsel++)
	{
		if (ENAREV)
		{
			printf("\nUnreverse sekvences for ");
			DisplayCompression(comsel);
			puts("\n");
		}

		int begin = StandaloneFirstByte(typcom[comsel]);
		for (index = input_size - 1; index >= begin; index--)
		{
			int length = buflen[index][comsel];
			if (length)
			{
				int offset = bufoff[index][comsel];
				index -= length - 1;
				buflen[index][comsel] = length;
				bufoff[index][comsel] = offset;

				if (ENAREV && index >= ENAZAC && index <= ENAEND)
					printf("Unreverse sekvence  %5u..%-5u  Len:%-5u  Off:%-5u \n",
						index, index + length - 1, length, offset);
			}
			else
			{
				int idxend = index;
				while (index >= begin && !buflen[index][comsel]) index--;
				int len = idxend - index;
				index++;
				bufoff[index][comsel] = 0;
				buflen[index][comsel] = len;
				bufoff[idxend][comsel] = 0;
				buflen[idxend][comsel] = len;

				if (ENAREV && index >= ENAZAC && index <= ENAEND)
					printf("Unreverse unpacked  %5u..%-5u  Len:%-5u \n",
						index, index + len - 1, len);
			}
		}
	}

	////////////////////////
	// Optimize sekvences //
	////////////////////////

	int optimized = 0;
	for (int comsel = 0; comsel < compressions; comsel++)
	{
		// Compute length with end marks //
		sumcom[comsel] = buftax[input_size - 1][comsel] + GetBitsEndMark(comsel);
		DisplayPackModel(comsel, input_size);
		if (ENAOPT && sumall[comsel] != sumcom[comsel])
		{
			printf("Error: Bad lengths  Lz: %-6u  Stat: %-6u  Dif: %+-5d  in  ", sumcom[comsel], sumall[comsel], sumall[comsel] - sumcom[comsel]);
			DisplayCompression(comsel);
			putchar('\n');
		}

		if (ENAOPT)
		{
			printf("\nOptimize sekvences for ");
			DisplayCompression(comsel);
			puts("\n");
		}

		int begin = StandaloneFirstByte(typcom[comsel]);
		int idxzac = begin;
		while (idxzac < input_size && bufoff[idxzac][comsel]) idxzac += buflen[idxzac][comsel];
		int idxend = idxzac + buflen[idxzac][comsel];
		while (idxzac > begin || idxend < input_size)
		{
			if (ENAOPT && idxend >= ENAZAC && idxzac <= ENAEND)
				printf("Work Unp % 5u..%-5u Len:%-5u\n", idxzac, idxend - 1, idxend - idxzac);

			int cont = 1;
			while (cont)
			{
				cont = 0;

				while (idxzac > begin)	// sek + unp = unp //
				{
					if (!bufoff[idxzac - 1][comsel]) printf("Error in optimization 1: No sekvence at %+d  Zac:%u End:%u\n", idxzac - 1, idxzac, idxend);

					int sekzac = idxzac - buflen[idxzac - 1][comsel];
					int sekend = idxzac;

					if (sekzac < begin) printf("Error in optimization 2: Bad sekvence at %d..%d\n", sekzac, sekend - 1);
					if (!OptimizeSekvence(comsel, sekzac, sekzac, sekend, idxend)) break;

					cont = 1;
					idxzac = sekzac;
					if (idxzac > begin && !bufoff[idxzac - 1][comsel]) // old unp + new unp => one unp //
					{
						int len1 = buflen[idxzac - 1][comsel];
						int len2 = buflen[idxzac - 0][comsel];
						int idxlen = len1 + len2;

						if (ENAOPT && idxend >= ENAZAC && idxzac <= ENAEND)
							printf("Join Unp % 5u..%-5u Len:%-5u + Unp % 5u..%-5u Len:%-5u => Unp % 5u..%-5u Len:%-5u\n",
								idxzac - len1, idxzac - 1, len1,
								idxzac, idxend - 1, len2,
								idxzac - len1, idxend - 1, idxlen);

						idxzac = idxzac - len1;
						buflen[idxzac - 0][comsel] = idxlen;
						buflen[idxend - 1][comsel] = idxlen;
					}
				}

				while (idxend < input_size)	// unp + sek = unp //
				{
					if (!bufoff[idxend][comsel]) printf("Error in optimization 3: No sekvence at %+d  Zac:%u End:%u\n", idxend, idxzac, idxend);

					int sekzac = idxend;
					int sekend = idxend + buflen[idxend][comsel];

					if (sekend > input_size) printf("Error in optimization 4: Bad sekvence at %d..%d\n", sekzac, sekend - 1);
					if (!OptimizeSekvence(comsel, idxzac, sekzac, sekend, sekend)) break;

					cont = 1;
					idxend = sekend;
					if (idxend < input_size && !bufoff[idxend][comsel]) // new unp + old unp => one unp //
					{
						int len1 = buflen[idxzac][comsel];
						int len2 = buflen[idxend][comsel];
						int idxlen = len1 + len2;

						if (ENAOPT && idxend >= ENAZAC && idxzac <= ENAEND)
							printf("Join Unp % 5u..%-5u Len:%-5u + Unp % 5u..%-5u Len:%-5u => Unp % 5u..%-5u Len:%-5u\n",
								idxzac, idxend - 1, len1,
								idxend, idxend + len2 - 1, len2,
								idxzac, idxend + len2 - 1, idxlen);

						idxend = idxend + len2;
						buflen[idxzac - 0][comsel] = idxlen;
						buflen[idxend - 1][comsel] = idxlen;
					}
				}
			}

			// previous unp + sek + this unp => unp //
			int prezac = idxzac;
			while (prezac > begin && bufoff[prezac - 1][comsel]) prezac -= buflen[prezac - 1][comsel];
			if (prezac > begin)
			{
				int preend = prezac;
				prezac -= buflen[prezac - 1][comsel];
				if (prezac < begin) printf("Error in optimization 5: Bad sekvence at %d..%d\n", prezac, preend - 1);
				if (OptimizeSekvence(comsel, prezac, preend, idxzac, idxend))
				{
					idxzac = prezac;
					continue;
				}
			}

			// this unp + sek + next unp => unp //
			int nxtend = idxend;
			while (nxtend < input_size && bufoff[nxtend][comsel]) nxtend += buflen[nxtend][comsel];
			if (nxtend < input_size)
			{
				int nxtzac = nxtend;
				nxtend += buflen[nxtend][comsel];
				if (nxtend > input_size) printf("Error in optimization 6: Bad sekvence at %d..%d\n", nxtzac, nxtend - 1);
				if (OptimizeSekvence(comsel, idxzac, idxend, nxtzac, nxtend))
				{
					idxend = nxtend;
					continue;
				}
				else
				{
					idxzac = nxtzac;
					idxend = nxtend;
					continue;
				}
			}
			break;
		}
		DisplayPackModel(comsel, input_size);
		if (sumall[comsel] != sumcom[comsel]) optimized++;

		sumdec[comsel] = depackers && declen[comsel] ? ((sumall[comsel] + 7) >> 3) + declen[comsel] : 0;
	}

	clock_t run_end = clock();

	//////////////////////
	// Find best result //
	//////////////////////

	int best_len_result = 99999999;
	int best_dec_result = 99999999;
	int best_len_compress = -1;
	int best_dec_compress = -1;
	for (int i = 0; i < compressions; i++)
	{
		if (best_len_result > sumall[i])
		{
			best_len_compress = i;
			best_len_result = sumall[i];
		}
		if (best_dec_result > sumdec[i] && declen[i])
		{
			best_dec_compress = i;
			best_dec_result = sumdec[i];
		}
	}

	if (ENAOPT && optimized)
	{
		printf("\nOptimized %u of %u compressions:\n\n", optimized, compressions);
		for (int i = 0; i < compressions; i++)
		{
			int bitlzc = sumcom[i];
			int bitend = sumall[i];
			DisplayCompression(i);
			printf("%5u   => %5u %6u", input_size, (bitlzc + 7) >> 3, bitlzc);
			if (bitend != bitlzc) printf("   => %5u %6u  optimized", (bitend + 7) >> 3, bitend);
			if (bitend == best_len_result) printf("   ... best result");
			putchar('\n');
		}
		putchar('\n');
	}

	////////////////
	// Statistics //
	////////////////

	if (show_statistics)
	{
		printf("\nCompression statistic:\n\n"
			"  Input length ... %5u bytes   %6u bits\n"
			"   Best result ... %5u bytes   %6u bits ... ",
			input_size, input_size << 3, (best_len_result + 7) >> 3, best_len_result);
		DisplayCompression(best_len_compress); putchar('\n');
		if (depackers && best_dec_compress >= 0)
		{
			printf(
				"   With depack ... %5u bytes   %6u bits ... ",
				best_dec_result, best_dec_result<<3);
			DisplayCompression(best_dec_compress); putchar('\n');
		}

		for (int s = 1; s >= (compressions == 1 ? 1:0); s--)
		{
			puts(s ? (compressions < 2 ? "\nResults:"
				: "\nResults for all used compressions:")
					: (depackers ? "\nResults sorted by length with depacker:"
						:"\nResults sorted by compressed length:"));

			printf("\nCompression                         NumSek Packed NoPck     Overhead     Packed length "); if (depackers) printf("  With depacker");
			printf("\n=================================== ====== ====== =====  ==============  =============="); if (depackers) printf("  =============");
			putchar('\n');

			int este = 1;
			int disp = 0;
			while (este)
			{
				este = 0;
				int pack = 999999;
				for (int i = 0; i < compressions; i++)
				{
					int bits = depackers ? sumdec[i] : sumall[i];
					if (s || (bits && bits == disp))
					{
						char bufove[16]; sprintf(bufove, "(%u)", sumove[i]);
						char buflen[16]; sprintf(buflen, "(%u)", sumall[i]);
						DisplayCompression(i);
						printf(" %5u  %5u %5u  %5u %8s  %5u %8s",
							sumsek[i], sumpck[i], sumunp[i], (sumove[i] + 7) >> 3, bufove, (sumall[i] + 7) >> 3, buflen);
						if (depackers && declen[i]) printf("  +%4u = %5u", declen[i], sumdec[i]);

						if (compressions > 1)
						{
							if (sumall[i] == best_len_result &&
								 sumdec[i] == best_dec_result) printf(" ... best compress and depacker");
							else if (sumall[i] == best_len_result) printf(" ... best compress");
							else if (sumdec[i] == best_dec_result) printf(" ... best depacker");
						}

						putchar('\n');
					}
					if (bits > disp && pack > bits)
					{
						pack = bits;
						este = 1;
					}
				}
				disp = pack;
				if (s) break;
			}
		}
		if (depackers && compressions > 1)
		{
			int sum = 0;
			for (int com = 0; com < compressions; com++) if (!declen[com]) sum++;
			if (sum) printf("\nNote: Sorted list does not contain %u compressions with not defined depacker length.\n", sum);
		}
		putchar('\n');

	}

	////////////////////////////
	// Generating output data //
	////////////////////////////

	if (!show_statistics && make_all_compressions && compressions > 1) putchar('\n');
	int best_of_best = depackers && best_dec_compress >= 0 ? best_dec_compress : best_len_compress;

	for (int comsel = make_all_compressions ? 0 : best_of_best;
			 comsel < (make_all_compressions ? compressions : best_of_best + 1);
			 comsel ++)
	{
		if (ENAOUT)
		{
			printf("\nGenerating output data for ");
			DisplayCompression(comsel);
			puts("\n");
		}

		index = 0;
		bits_stream = 0;
		packed_index = 0;
		packed_bytes = 0;
		unpack_bytes = 0;
		overhead_bits = 0;

		if (StandaloneFirstByte(typcom[comsel]))
		{
			if (ENAOUT && ENAZAC == 0) printf("Data     +0 0000    Packed     +0 0000    First byte: 0x%02X\n", *input_data);

			*packed_data = *input_data;
			packed_index++;
			unpack_bytes++;
			index++;
		}

		while (index < input_size)
		{
			if (ENAOUT && index >= ENAZAC && index <= ENAEND)
				printf("Data %+6d %04X    Packed %+6d %04X    ", index, index, packed_index, packed_index);

			int length = buflen[index][comsel];
			int offset = bufoff[index][comsel];
			if (offset)
			{
				if (ENAOUT && index >= ENAZAC && index <= ENAEND)
					printf("Sekvence Len: %-5u Off: %-5u Tax: %-2u   ", length, offset, GetTaxSekvence(comsel, offset, length));

				PutSekvence(comsel, offset, length);
				packed_bytes += length;
			}
			else
			{
				if (ENAOUT && index >= ENAZAC && index <= ENAEND)
					printf("Unpacked Len: %-5u            Tax: %-2u   ", length, GetTaxUnpack(comsel, length));

				PutUnpacked(comsel, index, length);
				unpack_bytes += length;
			}
			index += length;
			if (ENAOUT && index >= ENAZAC && index <= ENAEND) putchar('\n');
		}
		if (ENAOUT) printf("Data %+6d %04X    Packed %+6d %04X    End mark  ", index, index, packed_index, packed_index);
		PutEndMark(comsel);
		if (ENAOUT) putchar('\n');
		int bits_overhead = overhead_bits;
		if (bits_stream)
		{
			if (ENAOUT) printf("\nBitstream fill last byte ");
			while (bits_stream) PutBit(0);
		}
		if (ENAOUT) puts("\n");

		///////////////////////
		// Write output file //
		///////////////////////

		int txtlen = name_length + sprintf(name_out + name_length, "-t%c%c", typcom[comsel] & 0x0F | '0', typpos[comsel] & 0x0F | '0');
		int numofs = GetNumberOffsets(typpos[comsel]);
		if (numofs)
		{
			txtlen += sprintf(name_out + txtlen, "o%u", ofset1[comsel]);
			if (numofs > 1)
				txtlen += sprintf(name_out + txtlen, "o%u", ofset2[comsel]);
		}
		sprintf(name_out + txtlen, ".lzx");

		printf("Save file: %-*s  Length: %-5u Packed: %-5u NoPck: %-5u Overhead: %-5u (%u bits)\n",
			name_indent, name_out, packed_index, packed_bytes, unpack_bytes, (bits_overhead + 7) >> 3, bits_overhead);

		FILE *file_out = fopen(name_out, "wb"); if (!file_out) { perror(name_out); exit(4); }
		fwrite(packed_data, 1, packed_index, file_out);
		fclose(file_out);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////

	if (CLOCKS_PER_SEC == 1000)		// (should be defined as 1000 in time.h) //
	{
		int time_diff = run_end - run_begin;
		if (ENACOM || ENAINF || ENAOPT || time_diff > 0) printf("\nProcessing time: %u.%03u seconds   ", time_diff / 1000, time_diff % 1000);
	}
	if (ENACOM) printf("Compares: %u", cmps);
	putchar('\n');
	
	return 0;
}
